/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 1996, California Institute of Technology.  
// ALL RIGHTS RESERVED.   U.S. Government Sponsorship acknowledged. 
//
// Please read the full copyright notice in the file COPYRIGH 
// in this directory.
//
// Author: Todd Karakashian, NASA/Jet Propulsion Laboratory
//         Todd.K.Karakashian@jpl.nasa.gov
//
// $RCSfile: dodsutil.cc,v $ - Miscellaneous routines for DODS HDF server
//
// $Log: dodsutil.cc,v $
// Revision 1.3  1998/09/10 21:32:28  jehamby
// Properly escape high-ASCII strings with octstring() and hexstring()
//
// Revision 1.2  1998/02/05 20:14:30  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.1  1997/03/10 22:55:01  jimg
// New files for the 2.12 compatible HDF server
//
// Revision 1.3  1996/11/20  22:28:43  todd
// Modified to support UInt32 type.
//
// Revision 1.2  1996/10/07 21:15:17  todd
// Changes escape character to % from _.
//
// Revision 1.1  1996/09/24 22:38:16  todd
// Initial revision
//
//
/////////////////////////////////////////////////////////////////////////////

#include <strstream.h>
#include <iomanip.h>
#include <String.h>
#include "HDFStructure.h"
#include "HDFArray.h"
#include "dodsutil.h"
#include "dhdferr.h"
#include "hdf-dods.h"

String hexstring(unsigned char val) {
    static char buf[hdfdods::MAXSTR];

    ostrstream(buf,hdfdods::MAXSTR) << hex << setw(2) << setfill('0') << 
	(unsigned int)val << ends;

    return (String)buf;
}

char unhexstring(String s) {
    int val;
    static char buf[hdfdods::MAXSTR];

    strcpy(buf,(const char *)s);
    istrstream(buf,hdfdods::MAXSTR) >> hex >> val;

    return (char)val;
}

String octstring(unsigned char val) {
    static char buf[hdfdods::MAXSTR];

    ostrstream(buf,hdfdods::MAXSTR) << oct << setw(3) << setfill('0') << 
	(unsigned int)val << ends;

    return (String)buf;
}

char unoctstring(String s) {
    int val;
    static char buf[hdfdods::MAXSTR];

    strcpy(buf,(const char *)s);
    istrstream(buf,hdfdods::MAXSTR) >> oct >> val;

    return (char)val;
}


// replace characters that are not allowed in DODS identifiers
String id2dods(String s) {
    static Regex badregx("[^0-9a-zA-Z_%]");
    const char ESC = '%';

    int index;
    while ( (index = s.index(badregx)) >= 0)
	s.at(index,1) = (String)ESC + hexstring(s[index]);
    if (isdigit(s[0]))
	s.before(0) = '_';
    return s;
}

String dods2id(String s) {
    static Regex escregx("%[0-7][0-9a-fA-F]");

    int index;
    while ( (index = s.index(escregx)) >= 0)
	s.at(index,3) = unhexstring(s.at(index+1,2));

    return s;
}

// Escape non-printable characters and quotes from an HDF attribute
String escattr(String s) {
    static Regex nonprintable("[^ !-~]");
    const char ESC = '\\';
    const char QUOTE = '\"';
    const String ESCQUOTE = (String)ESC + (String)QUOTE;

    // escape non-printing characters with octal escape
    int index = 0;
    while ( (index = s.index(nonprintable)) >= 0)
	s.at(index,1) = (String)ESC + octstring(s[index]);

    // escape " with backslash
    index = 0;
    while ( (index = s.index(QUOTE,index)) >= 0) {
	s.at(index,1) = ESCQUOTE;
	index += ESCQUOTE.length();
    }
	   
    return s;
}

// Un-escape special characters, quotes and backslashes from an HDF attribute.
// Note: A regex to match one \ must be defined as
//          Regex foo = "\\\\";
//       because both C++ strings and libg++ regex's also employ \ as
//       an escape character!
String unescattr(String s) {
    static Regex escregx("\\\\[01][0-7][0-7]");  // matches 4 characters
    static Regex escquoteregex("[^\\\\]\\\\\"");  // matches 3 characters
    static Regex escescregex("\\\\\\\\");      // matches 2 characters
    const char ESC = '\\';
    const char QUOTE = '\"';
    const String ESCQUOTE = (String)ESC + (String)QUOTE;

    // unescape any octal-escaped ASCII characters
    int index = 0;
    while ( (index = s.index(escregx,index)) >= 0) {
	s.at(index,4) = unoctstring(s.at(index+1,3));
	index++;
    }

    // unescape any escaped quotes
    index = 0;
    while ( (index = s.index(escquoteregex,index)) >= 0) {
	s.at(index+1,2) = QUOTE;
	index++;
    }

    // unescape any escaped backslashes
    index = 0;
    while ( (index = s.index(escescregex,index)) >= 0) {
	s.at(index,2) = ESC;
	index++;
    }
    
    return s;
}

HDFStructure *CastBaseTypeToStructure(BaseType *p) {
    if (p->type_name() != "Structure")
	THROW(dhdferr_consist);
    return (HDFStructure *)p;
}

HDFArray *CastBaseTypeToArray(BaseType *p) {
    if (p->type_name() != "Array")
	THROW(dhdferr_consist);
    return (HDFArray *)p;
}

int *CastBaseTypeToInt(BaseType *p) {
    if (p->type_name() != "Int32")
	THROW(dhdferr_consist);
    return (int *)p;
}

double *CastBaseTypeToDouble(BaseType *p) {
    if (p->type_name() != "Float64")
	THROW(dhdferr_consist);
    return (double *)p;
}

